/* ----------------------------------------------------------------------- *
 *   Copyright 2017 Red Hat, Inc. All rights reserved.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ----------------------------------------------------------------------- */

#include <errno.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <linux/magic.h>
#include <linux/auto_fs4.h>

/* Unfortunately the dev ioctl include file is not included in the
 * kernel-headers package (it's in the kernel-devel package) and
 * that seems to cause include dependency problems, that's been
 * fixed in upstream commit 9b88ee0f3bb4c but will take time to
 * find it's way into kernel-headers packages. But since this is
 * meant to be a public interface using the include file directly
 * should be ok becuase backward compatibility must be maintained.
 */
#include "auto_dev-ioctl.h"
#include "autofs_ismountpoint.h"

#define CONTROL_DEVICE  "/dev/autofs"

#define DEV_IOCTL_IS_MOUNTED	0x0001
#define DEV_IOCTL_IS_AUTOFS	0x0002
#define DEV_IOCTL_IS_OTHER	0x0004

/* Use CLOEXEC flag if possible. */
static int cloexec_works = 0;

static void check_cloexec(int fd)
{
	if (cloexec_works == 0) {
		int fl = fcntl(fd, F_GETFD);
		if (fl != -1)
			cloexec_works = (fl & FD_CLOEXEC) ? 1 : -1;
	}
	if (cloexec_works > 0)
		return;
	fcntl(fd, F_SETFD, FD_CLOEXEC);
	return;
}

static int open_fd(const char *path, int flags)
{
	int fd;

#if defined(O_CLOEXEC) && defined(SOCK_CLOEXEC)
	if (cloexec_works != -1)
		flags |= O_CLOEXEC;
#endif
	fd = open(path, flags);
	if (fd == -1)
		return -1;
	check_cloexec(fd);
	return fd;
}

/* Get an fd to the autofs misc device node for our operations. */
static int get_autofs_dev_ioctl_fd(void)
{
	int devfd;

	devfd = open_fd(CONTROL_DEVICE, O_RDONLY);
	if (devfd == -1)
		return -1;
	else {
		struct autofs_dev_ioctl param;

		/* Check compile version against kernel.
		 *
		 * The requirement is the major version must be the
		 * same and the user space minor version must be
		 * less than or equal to the kernel space minor
		 * version.
		 *
		 * Selinux may allow us to open the device but not
		 * actually allow us to do anything.
		 */
		init_autofs_dev_ioctl(&param);
		if (ioctl(devfd, AUTOFS_DEV_IOCTL_VERSION, &param) == -1) {
			close(devfd);
			return -1;
		}

		/* Any misc device ioctl major version (currently 1) is ok
		 * but the minimum needed minor version will be 1 when the
		 * CAP_SYS_ADMIN check is fixed.
		 */
		if (param.ver_minor < 1) {
			close(devfd);
			return -1;
		}

		return devfd;
	}
}

/* Allocate a parameter struct for misc device ioctl which includes
 * a path. This is used when getting the last mount requester uid
 * and gid and when checking if a path within the autofs filesystem
 * is a mount point. The path is added to the end of the struct.
 */
static struct autofs_dev_ioctl *alloc_dev_ioctl_path(int ioctlfd, const char *path)
{
	struct autofs_dev_ioctl *ioctl;
	size_t size, p_len;

	if (!path) {
		errno = EINVAL;
		return NULL;
	}

	p_len = strlen(path);
	size = sizeof(struct autofs_dev_ioctl) + p_len + 1;
	ioctl = malloc(size);
	if (!ioctl)
		return NULL;

	init_autofs_dev_ioctl(ioctl);
	ioctl->ioctlfd = ioctlfd;
	ioctl->size = size;
	memcpy(ioctl->path, path, p_len);
	ioctl->path[p_len] = '\0';

	return ioctl;
}

static void free_dev_ioctl_path(struct autofs_dev_ioctl *ioctl)
{
	free(ioctl);
	return;
}

/*
 * Check if the given path is an autofs mountpoint.
 *
 * The path is considered a mountpoint if it is itself a mountpoint
 * or contains a mount, such as a multi-mount without a root mount.
 * In addition, if the path is itself a mountpoint we return whether
 * the mounted file system is an autofs filesystem or other file
 * system.
 */
unsigned int autofs_ismountpoint(const char *path)
{
	struct autofs_dev_ioctl *param;
	int misc_dev_fd;
	int err;
	int ret;

	if (!path)
		return AUTOFS_ISMOUNTPOINT_ERR_PATH;

	misc_dev_fd = get_autofs_dev_ioctl_fd();
	if (misc_dev_fd == -1)
		return AUTOFS_ISMOUNTPOINT_ERR_FD;

	/* Outside of autofs we probably don't want or need to use
	 * an autofs mount ioctl fd, just being able to get info.
	 * about the given path via the misc device ioctl should be
	 * enough for now.
	 */
	param = alloc_dev_ioctl_path(-1, path);
	if (!param) {
		close(misc_dev_fd);
		return AUTOFS_ISMOUNTPOINT_ERR_ALLOC;
	}

	/* Consider all autofs mount types, indirect, direct and offset */
	set_autofs_type_any(&param->ismountpoint.in.type);

	err = ioctl(misc_dev_fd, AUTOFS_DEV_IOCTL_ISMOUNTPOINT, param);
	if (err == -1) {
		int save_errno = errno;
		free_dev_ioctl_path(param);
		close(misc_dev_fd);
		errno = save_errno;
		return AUTOFS_ISMOUNTPOINT_ERR_IOCTL;
	}

	ret = AUTOFS_ISMOUNTPOINT_NOT_AUTOFS;
	if (err) {
		ret = AUTOFS_ISMOUNTPOINT_EXISTS;

		if (param->ismountpoint.out.magic == AUTOFS_SUPER_MAGIC)
			ret |= AUTOFS_ISMOUNTPOINT_AUTOFS;
		else
			ret |= AUTOFS_ISMOUNTPOINT_OTHER;
	}

	free_dev_ioctl_path(param);
	close(misc_dev_fd);

	return ret;
}
